@@ -5,12 +5,14 @@ from PyQt5.QtCore import QSize
 from PyQt5.QtGui import QFont
 from coolname import generate_slug
 from ...factory import list_algo, list_env, get_algo, get_env
-from ...utils import ystring, load_config, config_list_to_dict, normalize_routine
+from ...utils import ystring, load_config, config_list_to_dict, normalize_routine, instantiate_env
 from ..components.variable_item import variable_item
 from ..components.objective_item import objective_item
 from ..components.constraint_item import constraint_item
 from ..windows.review_dialog import BadgerReviewDialog
 from ..windows.opt_monitor import BadgerOptMonitor
+from ..windows.var_dialog import BadgerVariableDialog
+from ..windows.edit_script_dialog import BadgerEditScriptDialog
 
 
 CONS_RELATION_DICT = {
@@ -28,6 +30,8 @@ class BadgerRoutinePage(QWidget):
 
         self.algos = list_algo()
         self.envs = list_env()
+        self.env = None
+        self.script = ''
 
         self.init_ui()
         self.config_logic()
@@ -43,11 +47,22 @@ class BadgerRoutinePage(QWidget):
 
         group_algo = QGroupBox('Algorithm')
         vbox_algo = QVBoxLayout(group_algo)
+        action_algo = QWidget()
+        hbox_action_algo = QHBoxLayout(action_algo)
+        hbox_action_algo.setContentsMargins(0, 0, 0, 0)
         self.cb_algo = cb_algo = QComboBox()
         cb_algo.setItemDelegate(QStyledItemDelegate())
         cb_algo.addItems(self.algos)
         cb_algo.setCurrentIndex(-1)
-        vbox_algo.addWidget(cb_algo)
+        self.check_use_script = check_use_script = QCheckBox('Use Script')
+        check_use_script.setChecked(False)
+        self.btn_edit_script = btn_edit_script = QPushButton('Edit Script')
+        btn_edit_script.setFixedSize(96, 24)
+        btn_edit_script.hide()
+        hbox_action_algo.addWidget(cb_algo, 1)
+        hbox_action_algo.addWidget(check_use_script)
+        hbox_action_algo.addWidget(btn_edit_script)
+        vbox_algo.addWidget(action_algo)
         self.edit_algo = edit_algo = QPlainTextEdit()
         edit_algo.setFixedHeight(128)
         vbox_algo.addWidget(edit_algo)
@@ -55,11 +70,19 @@ class BadgerRoutinePage(QWidget):
 
         group_env = QGroupBox('Environment')
         vbox_env = QVBoxLayout(group_env)
+        action_env = QWidget()
+        hbox_action_env = QHBoxLayout(action_env)
+        hbox_action_env.setContentsMargins(0, 0, 0, 0)
         self.cb_env = cb_env = QComboBox()
         cb_env.setItemDelegate(QStyledItemDelegate())
         cb_env.addItems(self.envs)
         cb_env.setCurrentIndex(-1)
-        vbox_env.addWidget(cb_env)
+        self.btn_env_play = btn_env_play = QPushButton('Playground')
+        btn_env_play.setFixedSize(96, 24)
+        btn_env_play.hide()  # hide for now to prevent confusions
+        hbox_action_env.addWidget(cb_env, 1)
+        hbox_action_env.addWidget(btn_env_play)
+        vbox_env.addWidget(action_env)
         self.edit_env = edit_env = QPlainTextEdit()
         edit_env.setFixedHeight(128)
         vbox_env.addWidget(edit_env)
@@ -79,10 +102,14 @@ class BadgerRoutinePage(QWidget):
         self.btn_un_all_var = btn_un_all_var = QPushButton('Uncheck All')
         btn_all_var.setFixedSize(96, 24)
         btn_un_all_var.setFixedSize(96, 24)
+        self.btn_add_var = btn_add_var = QPushButton('Add')
+        btn_add_var.setFixedSize(96, 24)
+        btn_add_var.setDisabled(True)
         self.check_only_var = check_only_var = QCheckBox('Checked Only')
         check_only_var.setChecked(False)
         hbox_action_var.addWidget(btn_all_var)
         hbox_action_var.addWidget(btn_un_all_var)
+        hbox_action_var.addWidget(btn_add_var)
         hbox_action_var.addStretch()
         hbox_action_var.addWidget(check_only_var)
         vbox_var.addWidget(action_var)
@@ -119,6 +146,7 @@ class BadgerRoutinePage(QWidget):
         hbox_action_con.setContentsMargins(0, 0, 0, 0)
         self.btn_add_con = btn_add_con = QPushButton('Add')
         btn_add_con.setFixedSize(96, 24)
+        btn_add_con.setDisabled(True)
         hbox_action_con.addWidget(btn_add_con)
         hbox_action_con.addStretch()
         vbox_con.addWidget(action_con)
@@ -172,9 +200,13 @@ class BadgerRoutinePage(QWidget):
     def config_logic(self):
         self.btn_back.clicked.connect(self.go_home)
         self.cb_algo.currentIndexChanged.connect(self.select_algo)
+        self.check_use_script.stateChanged.connect(self.toggle_use_script)
+        self.btn_edit_script.clicked.connect(self.edit_script)
         self.cb_env.currentIndexChanged.connect(self.select_env)
+        self.btn_env_play.clicked.connect(self.open_playground)
         self.btn_all_var.clicked.connect(self.check_all_var)
         self.btn_un_all_var.clicked.connect(self.uncheck_all_var)
+        self.btn_add_var.clicked.connect(self.add_var)
         self.btn_all_obj.clicked.connect(self.check_all_obj)
         self.btn_un_all_obj.clicked.connect(self.uncheck_all_obj)
         self.check_only_var.stateChanged.connect(self.toggle_check_only_var)
@@ -185,6 +217,8 @@ class BadgerRoutinePage(QWidget):
         self.btn_run.clicked.connect(self.run)
 
     def refresh_ui(self, routine):
+        self.routine = routine  # save routine for future reference
+
         self.algos = list_algo()
         self.envs = list_env()
         # Clean up the constraints list
@@ -212,6 +246,10 @@ class BadgerRoutinePage(QWidget):
         idx_algo = self.algos.index(name_algo)
         self.cb_algo.setCurrentIndex(idx_algo)
         self.edit_algo.setPlainText(ystring(routine['algo_params']))
+        try:
+            self.script = routine['config']['script']
+        except KeyError:
+            self.script = None
 
         name_env = routine['env']
         idx_env = self.envs.index(name_env)
@@ -252,9 +290,15 @@ class BadgerRoutinePage(QWidget):
         if routine['config']['constraints'] is not None:
             for i in range(len(routine['config']['constraints'])):
                 name = next(iter(routine['config']['constraints'][i]))
-                relation, thres = routine['config']['constraints'][i][name]
-                relation = ['GREATER_THAN', 'LESS_THAN', 'EQUAL_TO'].index(relation)
-                self.add_constraint(name, relation, thres)
+                relation, thres = routine['config']['constraints'][i][name][:2]
+                try:
+                    routine['config']['constraints'][i][name][2]
+                    critical = True
+                except:
+                    critical = False
+                relation = ['GREATER_THAN', 'LESS_THAN',
+                            'EQUAL_TO'].index(relation)
+                self.add_constraint(name, relation, thres, critical)
 
         # Config the save settings
         name = routine['name']
@@ -262,7 +306,14 @@ class BadgerRoutinePage(QWidget):
         self.edit_save.setText(name)
         self.check_save.setChecked(False)
 
+        if self.script:
+            self.check_use_script.setChecked(True)
+
     def select_algo(self, i):
+        # Reset the script
+        self.script = ''
+        self.check_use_script.setChecked(False)
+
         if i == -1:
             self.edit_algo.setPlainText('')
             return
@@ -275,6 +326,53 @@ class BadgerRoutinePage(QWidget):
             self.cb_algo.setCurrentIndex(-1)
             return QMessageBox.critical(self, 'Error!', str(e))
 
+    def toggle_use_script(self):
+        if self.check_use_script.isChecked():
+            self.btn_edit_script.show()
+            self.edit_algo.setReadOnly(True)
+            self.refresh_params_algo()
+        else:
+            self.btn_edit_script.hide()
+            self.edit_algo.setReadOnly(False)
+
+    def edit_script(self):
+        algo = self.cb_algo.currentText()
+        dlg = BadgerEditScriptDialog(self, algo, self.script, self.script_updated)
+        dlg.exec()
+
+    def script_updated(self, text):
+        self.script = text
+        self.refresh_params_algo()
+
+    def refresh_params_algo(self):
+        if not self.script:
+            return
+
+        try:
+            tmp = {}
+            exec(self.script, tmp)
+            try:
+                tmp['generate']  # test if generate function is defined
+            except Exception as e:
+                QMessageBox.warning(self, 'Please define a valid generate function!', str(e))
+                return
+
+            env_params = load_config(self.edit_env.toPlainText())
+            try:
+                intf_name = self.configs['interface'][0]
+            except KeyError:
+                intf_name = None
+            configs = {
+                'params': env_params,
+                'interface': [intf_name]
+            }
+            env = instantiate_env(self.env, configs)
+            # Function generate comes from the script
+            params_algo = tmp['generate'](env, None)
+            self.edit_algo.setPlainText(ystring(params_algo))
+        except Exception as e:
+            QMessageBox.warning(self, 'Invalid script!', str(e))
+
     def select_env(self, i):
         if i == -1:
             self.edit_env.setPlainText('')
@@ -283,21 +381,48 @@ class BadgerRoutinePage(QWidget):
             self.list_obj.clear()
             self.dict_obj.clear()
             self.configs = None
+            self.env = None
+            self.btn_add_con.setDisabled(True)
+            self.btn_add_var.setDisabled(True)
+            self.routine = None
             return
 
         name = self.envs[i]
         try:
-            _, configs = get_env(name)
+            env, configs = get_env(name)
             self.configs = configs
+            self.env = env
+            self.btn_add_con.setDisabled(False)
+            self.btn_add_var.setDisabled(False)
+            if self.check_use_script.isChecked():
+                self.refresh_params_algo()
         except Exception as e:
             self.configs = None
+            self.env = None
             self.cb_env.setCurrentIndex(-1)
+            self.btn_add_con.setDisabled(True)
+            self.btn_add_var.setDisabled(True)
+            self.routine = None
             return QMessageBox.critical(self, 'Error!', str(e))
 
         self.edit_env.setPlainText(ystring(configs['params']))
 
         self.list_var.clear()
-        for var in configs['variables']:
+        vars_env = configs['variables']
+        vars_combine = [*vars_env]
+        if self.routine:  # check for the temp variables in vocs
+            vars_vocs = self.routine['config']['variables']
+            var_names_vocs = [next(iter(var)) for var in vars_vocs]
+            var_names_env = [next(iter(var)) for var in vars_env]
+            env_instance = instantiate_env(env, configs)
+            for name in var_names_vocs:
+                if name in var_names_env:
+                    continue
+
+                _var = {}
+                _var[name] = env_instance._get_vrange(name)
+                vars_combine.append(_var)
+        for var in vars_combine:
             name = next(iter(var))
             vrange = var[name]
             item = QListWidgetItem(self.list_var)
@@ -317,6 +442,10 @@ class BadgerRoutinePage(QWidget):
             self.dict_obj[obj] = item
 
         self.list_con.clear()
+        self.routine = None
+
+    def open_playground(self):
+        pass
 
     def check_all_var(self):
         for i in range(self.list_var.count()):
@@ -330,6 +459,43 @@ class BadgerRoutinePage(QWidget):
             item_widget = self.list_var.itemWidget(item)
             item_widget.check_name.setChecked(False)
 
+    def add_var(self):
+        # TODO: Use a cached env
+        env_params = load_config(self.edit_env.toPlainText())
+        try:
+            intf_name = self.configs['interface'][0]
+        except KeyError:
+            intf_name = None
+        configs = {
+            'params': env_params,
+            'interface': [intf_name]
+        }
+
+        dlg = BadgerVariableDialog(self, self.env, configs, self.add_var_to_list)
+        dlg.exec()
+
+    def add_var_to_list(self, name, min, max):
+        # Check if already in the list
+        ok = False
+        try:
+            self.dict_var[name]
+        except:
+            ok = True
+
+        if not ok:
+            QMessageBox.warning(self, 'Variable already exists!', f'Variable {name} already exists!')
+            return 1
+
+        item = QListWidgetItem(self.list_var)
+        var_item = variable_item(name, [min, max], self.toggle_var)
+        var_item.check_name.setChecked(True)
+        item.setSizeHint(var_item.sizeHint())
+        self.list_var.addItem(item)
+        self.list_var.setItemWidget(item, var_item)
+        self.dict_var[name] = item
+
+        return 0
+
     def check_all_obj(self):
         for i in range(self.list_obj.count()):
             item = self.list_obj.item(i)
@@ -400,15 +566,16 @@ class BadgerRoutinePage(QWidget):
         else:
             self.edit_save.setDisabled(True)
 
-    def add_constraint(self, name=None, relation=0, threshold=0):
+    def add_constraint(self, name=None, relation=0, threshold=0, critical=False):
         if self.configs is None:
             return
 
         options = self.configs['observations']
         item = QListWidgetItem(self.list_con)
         con_item = constraint_item(options,
-            lambda: self.list_con.takeItem(self.list_con.row(item)),
-            name, relation, threshold)
+                                   lambda: self.list_con.takeItem(
+                                       self.list_con.row(item)),
+                                   name, relation, threshold, critical)
         item.setSizeHint(con_item.sizeHint())
         self.list_con.addItem(item)
         self.list_con.setItemWidget(item, con_item)
@@ -449,11 +616,14 @@ class BadgerRoutinePage(QWidget):
         for i in range(self.list_con.count()):
             item = self.list_con.item(i)
             item_widget = self.list_con.itemWidget(item)
+            critical = item_widget.check_crit.isChecked()
             con_name = item_widget.cb_obs.currentText()
             relation = CONS_RELATION_DICT[item_widget.cb_rel.currentText()]
             value = item_widget.sb.value()
             _dict = {}
             _dict[con_name] = [relation, value]
+            if critical:
+                _dict[con_name].append('CRITICAL')
             constraints.append(_dict)
 
         configs = {
@@ -461,6 +631,8 @@ class BadgerRoutinePage(QWidget):
             'objectives': objectives,
             'constraints': constraints,
         }
+        if self.check_use_script.isChecked():
+            configs['script'] = self.script
 
         routine = {
             'name': name,
